"""The names :py:class:`.render.Color`, :py:class:`.render.Material`,
:py:class:`.render.Renderer`, :py:class:`.wrapper.NTracer`,
:py:data:`.wrapper.CUBE`, :py:data:`.wrapper.SPHERE` and
:py:func:`.kdtree_builder.build_composite_scene` are imported here for
convenience."""


from ntracer.render import Color,Material,Renderer
from ntracer.wrapper import NTracer,CUBE,SPHERE
from ntracer.kdtree_builder import build_composite_scene




#!/usr/bin/python

from __future__ import print_function


import sys
import itertools
import functools
import argparse
import pygame
from ntracer import NTracer, Renderer, build_composite_scene
from ntracer import wavefront_obj


MOVE_SENSITIVITY = 0.01
ROTATE_SENSITIVITY = 0.002

UI_FOREGROUND = 100,100,100
UI_BACKGROUND = 255,255,255
UI_BASE_SIZE = 15
UI_TEXT_SPACE = 80
UI_ALPHA_NORMAL = 128
UI_ALPHA_HOVER = 200


O_LEFT = 0
O_RIGHT = 1
O_UP = 2
O_DOWN = 3

M_NORMAL = 0
M_HOVER = 1
M_PRESSED = 2

SLIDER_EVENT = pygame.USEREVENT + 1
SLIDER_DELAY = 50


def make_arrowhead(size,orient):
    surf = pygame.Surface(size)
    surf.fill(UI_BACKGROUND)
    surf.lock()
    try:
        area = pygame.Rect(1,1,size[0]-2,size[1]-2)
        pygame.draw.rect(surf,UI_FOREGROUND,area,1)
        area.inflate_ip(-5,-5)
        points = [
            [area.midleft,area.topright,area.bottomright], # O_LEFT
            [area.bottomleft,area.topleft,area.midright], # O_RIGHT
            [area.bottomleft,area.midtop,area.bottomright], # O_UP
            [area.topleft,area.midbottom,area.topright] # O_DOWN
            ][orient]
        pygame.draw.aalines(surf,UI_FOREGROUND,True,points)
    finally:
        surf.unlock()

    return surf


class Slider:
    arrows = [None] * 4

    @classmethod
    def arrow(cls,orient):
        if not cls.arrows[orient]:
            cls.arrows[orient] = make_arrowhead((UI_BASE_SIZE,UI_BASE_SIZE),orient)
        return cls.arrows[orient]

    def __init__(self,text,pos,on_slide):
        self.pos = pos
        self.text = UI.make_text_block(text,(UI_TEXT_SPACE - 2,UI_BASE_SIZE))
        self.text.set_alpha(UI_ALPHA_NORMAL)
        self.on_slide = on_slide

    def draw(self):
        screen.blit(self.text,(self.pos[0] + UI_BASE_SIZE + 1,self.pos[1]))


class SliderButton:
    def __init__(self,slider):
        self.slider = slider

    @property
    def area(self):
        return pygame.Rect(self.pos,(UI_BASE_SIZE,UI_BASE_SIZE))

    def draw(self,mode):
        a = Slider.arrow(self.orient)
        alpha = UI_ALPHA_NORMAL
        flags = 0
        if mode == M_PRESSED:
            alpha = 255
            flags = pygame.BLEND_MULT
        elif mode == M_HOVER:
            alpha = UI_ALPHA_HOVER

        a.set_alpha(alpha)
        screen.blit(a,self.pos,special_flags=flags)

    def on_press(self):
        UI.begin_tick()

    def on_release(self):
        UI.tick()
        UI.end_tick()

class LeftSliderButton(SliderButton):
    orient = O_LEFT

    @property
    def pos(self):
        return self.slider.pos

    def on_tick(self,t):
        self.slider.on_slide(-t)


class RightSliderButton(SliderButton):
    orient = O_RIGHT

    @property
    def pos(self):
        return (self.slider.pos[0] + UI_BASE_SIZE + UI_TEXT_SPACE,self.slider.pos[1])

    def on_tick(self,t):
        self.slider.on_slide(t)


class Label:
    def __init__(self,text,pos):
        self.text = UI.make_text_block(text,(UI_BASE_SIZE*2 + 2 + UI_TEXT_SPACE,UI_BASE_SIZE))
        self.text.set_alpha(UI_ALPHA_NORMAL)
        self.pos = pos

    def draw(self):
        screen.blit(self.text,self.pos)


class UI:
    buttons = []
    components = []
    active = None
    hover = False
    pressed = False
    lasttime = None

    _font = None

    @classmethod
    def font(cls):
        if not cls._font:
            cls._font = pygame.font.SysFont('None',UI_BASE_SIZE)
        return cls._font

    @classmethod
    def make_text_block(cls,text,size):
        font = cls.font()
        fsurf = font.render(text,True,UI_FOREGROUND,UI_BACKGROUND)
        block = pygame.Surface(size)
        block.fill(UI_BACKGROUND)
        block.blit(
            fsurf,
            (   (size[0] - fsurf.get_width()) // 2,
                (size[1] - font.get_ascent()) // 2))
        return block

    @classmethod
    def on_mousemove(cls,e):
        if cls.active:
            hover = cls.active.area.collidepoint(e.pos)
            if cls.hover != hover:
                cls.hover = hover
                cls.redraw_active()
            if cls.hover or cls.pressed:
                return

        cls.active = None
        for c in cls.buttons:
            if c.area.collidepoint(e.pos):
                cls.active = c
                cls.hover = True
                cls.redraw_active()
                break

    @classmethod
    def on_mousedown(cls,e):
        if e.button == 1 and cls.active and not cls.pressed:
            cls.pressed = True
            cls.redraw_active()
            cls.active.on_press()

    @classmethod
    def on_mouseup(cls,e):
        if e.button == 1 and cls.active and cls.pressed:
            cls.pressed = False
            cls.redraw_active()
            cls.active.on_release()
            

    @classmethod
    def on_active(cls,e):
        if (not e.gain) and cls.active and cls.pressed:
            cls.pressed = False
            cls.redraw_active()
            cls.active.on_release()

    @classmethod
    def full_draw(cls):
        for c in cls.components:
            c.draw()

        for b in cls.buttons:
            mode = M_NORMAL
            if cls.active == b:
                mode = M_HOVER
                if cls.pressed:
                    mode = M_PRESSED
            b.draw(mode)

    @classmethod
    def redraw_active(cls):
        area = cls.active.area
        screen.blit(current_img,area,area)
        mode = M_NORMAL
        if cls.pressed: mode = M_PRESSED
        elif cls.hover: mode = M_HOVER
        cls.active.draw(mode)
        pygame.display.update(area)

    @classmethod
    def begin_tick(cls):
        # set_timer has the side-effect of initializing the timer subsystem,
        # which get_ticks doesn't do
        pygame.time.set_timer(SLIDER_EVENT,SLIDER_DELAY)

        cls.lasttime = pygame.time.get_ticks()

    @classmethod
    def end_tick(cls):
        cls.lasttime = None
        pygame.time.set_timer(SLIDER_EVENT,0)
        pygame.event.clear(SLIDER_EVENT)

    @classmethod
    def ticking(cls):
        return cls.lasttime is not None

    @classmethod
    def tick(cls,reset=False):
        t = pygame.time.get_ticks()
        cls.active.on_tick(cls.lasttime - t)
        cls.lasttime = t

        if reset:
            pygame.event.clear(SLIDER_EVENT)
            pygame.time.set_timer(SLIDER_EVENT,SLIDER_DELAY)

    @classmethod
    def slider(cls,text,pos,on_slide):
        s = Slider(text,pos,on_slide)
        cls.components.append(s)
        cls.buttons.append(LeftSliderButton(s))
        cls.buttons.append(RightSliderButton(s))

    @classmethod
    def label(cls,text,pos):
        cls.components.append(Label(text,pos))


def d_symbol(d):
    return 'XYZ'[d] if d < 3 else 'D' + str(d+1)


def excepthook(type,value,traceback):
    if isinstance(value,Exception):
        print(str(value),file=sys.stderr)
    else:
        sys.__excepthook__(type,value,traceback)
    
sys.excepthook = excepthook

argp = argparse.ArgumentParser(description='''Navigate around a hypercube or a
mesh from a file. By default, a 3-dimensional cube will be displayed. To
display a hypercube, use --dimension to specify the dimensionality.
Alternatively, an arbitrary 3D mesh can be specified with --file, which should
specify the path to a wavefront obj file. When using a file, only the polygon
data will displayed. Lines and curves will be ignored.''')
argp.add_argument('-d','--dimension',default=3,type=int,metavar='N',help='the dimensionality of the hypercube and the navigable space.')
argp.add_argument('-f','--file',metavar='PATH',help='a wavefront obj file to load and display')
argp.add_argument('--debug',action='store_true',help="don't catch exceptions from errors")
args = argp.parse_args()

if args.debug:
    sys.excepthook = sys.__excepthook__

ntracer = NTracer(args.dimension)

if args.file:
    if ntracer.dimension != 3:
        sys.exit('a file can only be displayed with a dimensionality of 3')

    try:
        triangles = wavefront_obj.load_obj(args.file)
    except Exception as e:
        if args.debug:
            raise
        sys.exit('could not load file: ' + str(e))
        
    scene = build_composite_scene(ntracer,triangles)
else:
    scene = ntracer.BoxScene()

pygame.display.init()
pygame.font.init()

pygame.display.set_caption('ntracer')
screen = pygame.display.set_mode((640,480),pygame.RESIZABLE)

current_img = pygame.Surface(screen.get_size())
next_img = pygame.Surface(screen.get_size())

started = False

def translate(d):
    def inner(t):
        global started
        camera.origin += camera.axes[d] * (t * MOVE_SENSITIVITY)
        if not started:
            begin_render()
            started = True
    return inner

UI.label('Slide',(15,15))
for i in range(ntracer.dimension):
    UI.slider(d_symbol(i),(15,35+20*i),translate(i))


def rotate(d1,d2):
    def inner(t):
        global started
        m = ntracer.Matrix.rotation(camera.axes[d1],camera.axes[d2],t * ROTATE_SENSITIVITY)
        camera.transform(m)
        camera.normalize()

        if not started:
            begin_render()
            started = True
    return inner

UI.label('Turn',(15,50+20*ntracer.dimension))
for i,dd in enumerate(itertools.combinations(range(ntracer.dimension),2)):
    UI.slider(
        '{0} -> {1}'.format(*map(d_symbol,dd)),
        (15,70+20*(i+ntracer.dimension)),
        rotate(*dd))


camera = ntracer.Camera()
camera.translate(ntracer.Vector.axis(2,-5))
r = Renderer()

def begin_render():
    scene.set_camera(camera)
    r.begin_render(next_img,scene)

begin_render()

while True:
    e = pygame.event.wait()    
    if e.type == pygame.MOUSEMOTION:
        UI.on_mousemove(e)
    elif e.type == pygame.MOUSEBUTTONDOWN:
        UI.on_mousedown(e)
    elif e.type == pygame.MOUSEBUTTONUP:
        UI.on_mouseup(e)
    elif e.type == SLIDER_EVENT:
        UI.tick()
    elif e.type == pygame.USEREVENT:
        current_img,next_img = next_img,current_img
        screen.blit(current_img,(0,0))
        UI.full_draw()
        pygame.display.flip()
        if UI.ticking():
            UI.tick(True)
            begin_render()
        else:
            started = False
    elif e.type == pygame.VIDEORESIZE:
        r.abort_render()
        pygame.event.clear(pygame.USEREVENT);
        screen = pygame.display.set_mode(e.size,pygame.RESIZABLE)
        
        # the surfaces can be very large, so we free the memory first
        del current_img
        del next_img
        current_img = pygame.Surface(e.size)
        next_img = pygame.Surface(e.size)
        begin_render()
    elif e.type == pygame.QUIT:
        break
    


import sys
import os.path

from distutils import sysconfig
from distutils.core import setup,Command
from distutils.extension import Extension
from distutils.command.build_ext import build_ext
from distutils.errors import DistutilsSetupError
from distutils.dir_util import mkpath
from distutils import log
from distutils.dep_util import newer

try:
    from distutils.command.build_py import build_py_2to3 as build_py
except ImportError:
    from distutils.command.build_py import build_py

base_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0,os.path.join(base_dir,'support'))

import version
import generate_simd


GCC_OPTIMIZE_COMPILE_ARGS = [
    '-O3',
    '-fmerge-all-constants',
    '-funsafe-loop-optimizations',
    '-ffast-math',
    '-fsched2-use-superblocks',
    '-fstrict-enums',
    '-fno-enforce-eh-specs',
    '-fnothrow-opt',
    '-march=native',
#    '-DPROFILE_CODE',
    '--param','large-function-growth=500',
    '--param','inline-unit-growth=1000']
GCC_EXTRA_COMPILE_ARGS = ['-std=c++11','-fvisibility=hidden','-Wno-format','-Wno-invalid-offsetof']

# At the time of this writing, no version of MSVC has enough C++11 support to
# compile this package
MSVC_OPTIMIZE_COMPILE_ARGS = ['/Ox','/fp:fast','/volatile:iso','/Za']

DEFAULT_OPTIMIZED_DIMENSIONS = frozenset(range(3,9))


TRACER_TEMPLATE = """
#include <Python.h>
#include "fixed_geometry.hpp"

typedef fixed::item_store<{0},real> module_store;

#define MODULE_NAME tracer{0}
#include "ntracer_body.hpp"
"""


def check_min(x):
    if x < 3:
        raise DistutilsSetupError('dimension cannot be less than 3')
        
def parse_int(x):
    try:
        return int(x,10)
    except ValueError:
        raise DistutilsSetupError('invalid format for ranges')

def parse_ranges(x):
    items = set()
    parts = x.split(',')
    for p in parts:
        if (not p) or p.isspace(): continue
        p = p.partition('-')
        start = parse_int(p[0])
        check_min(start)
        end = p[2].strip()
        if end:
            end = parse_int(end)
            check_min(end)
            if end < start: start,end = end,start
            items.update(xrange(start,end+1))
        else:
            items.add(start)
            
    return items


class CustomBuildExt(build_ext):
    user_options = build_ext.user_options + [
        ('optimize-dimensions=',None,
         'which dimensionalities will have optimized versions of the main tracer module')]
    
    def initialize_options(self):
        build_ext.initialize_options(self)
        self.optimize_dimensions = None
        
    def special_tracer_file(self,n):
        return os.path.join(self.build_temp,'tracer{0}.cpp'.format(n))
        
    def special_tracer_ext(self,n):
        n = str(n)
        return Extension(
            'tracer'+n,
            [self.special_tracer_file(n),'src/py_common.cpp','src/simd.cpp'],
            depends=['src/simd.hpp.in','src/ntracer_body.hpp',
                'src/py_common.hpp','src/pyobject.hpp','src/geometry.hpp',
                'src/fixed_geometry.hpp','src/tracer.hpp','src/light.hpp',
                'src/render.hpp','src/camera.hpp','src/compatibility.hpp',
                'src/v_array.hpp','src/instrumentation.hpp'],
            include_dirs=['src'],
            libraries=['SDL'])
    
    def find_sdl(self,path):
        if os.path.isfile(os.path.join(path,'SDL.h')):
            return path
        
        path = os.path.join(path,'SDL')
        if os.path.isfile(os.path.join(path,'SDL.h')):
            return path
        
        return None
    
    def finalize_options(self):
        build_ext.finalize_options(self)
        
        self.optimize_dimensions = (parse_ranges(self.optimize_dimensions)
            if self.optimize_dimensions is not None
            else DEFAULT_OPTIMIZED_DIMENSIONS)
        
        for d in self.optimize_dimensions:
            self.extensions.append(self.special_tracer_ext(d))
        
        # try to find the SDL include directory
        for p in self.include_dirs:
            sp = self.find_sdl(p)
            if sp is not None:
                if sp is not p:
                    self.include_dirs.append(sp)
                break
        else:
            for p in ['/usr/include','/usr/local/include']:
                sp = self.find_sdl(p)
                if sp is not None:
                    self.include_dirs.append(sp)
                    break
            else:
                raise DistutilsSetupError(
                    'Unable to locate "SDL.h". Use --include-dirs to specify the location of the SDL header files.')
        
        # needed for simd.hpp (derived from simd.hpp.in)
        self.include_dirs.append(self.build_temp)
        self.include_dirs.append('src')

    def add_optimization(self):
        c = self.compiler.compiler_type
        if c == 'msvc':
            for e in self.extensions:
                e.extra_compile_args = MSVC_OPTIMIZE_COMPILE_ARGS
            return True
        if c == 'unix' or c == 'cygwin':
            cc = os.path.basename(sysconfig.get_config_var('CXX'))
            if 'gcc' in cc or 'g++' in cc or 'clang' in cc:
                for e in self.extensions:
                    e.extra_compile_args = GCC_EXTRA_COMPILE_ARGS[:]
                    e.extra_compile_args += GCC_OPTIMIZE_COMPILE_ARGS
                    if not self.debug:
                        e.extra_link_args = ['-s']
                return True
            
        return False

    def build_extensions(self):
        if not self.add_optimization():
            self.warn("don't know how to set optimization flags for this compiler")
        
        mkpath(self.build_temp,dry_run=self.dry_run)
        if self.optimize_dimensions:
            for d in self.optimize_dimensions:
                f = self.special_tracer_file(d)
                if self.force or not os.path.exists(f):
                    log.info('creating {0} from template'.format(f))
                    if not self.dry_run:
                        with open(f,'w') as out:
                            out.write(TRACER_TEMPLATE.format(d))
        
        simd_out = os.path.join(self.build_temp,'simd.hpp')
        simd_in = os.path.join(base_dir,'src','simd.hpp.in')
        if self.force or newer(simd_in,simd_out):
            log.info('creating {0} from {1}'.format(simd_out,simd_in))
            if not self.dry_run:
                generate_simd.generate(
                    os.path.join(base_dir,'src','intrinsics.json'),
                    simd_in,
                    simd_out)
        
        build_ext.build_extensions(self)


long_description = open('README.rst').read()
long_description = long_description[0:long_description.find('\n\n\n')]


setup(name='ntracer',
    author='Rouslan Korneychuk',
    author_email='rouslank@msn.com',
    url='https://github.com/Rouslan/NTracer',
    version=version.get_version(base_dir) or 'unversioned',
    packages=['ntracer','ntracer.tests'],
    scripts=['scripts/hypercube.py','scripts/polytope.py'],
    ext_package='ntracer',
    ext_modules=[
        Extension(
            'render',
            ['src/render.cpp','src/py_common.cpp'],
            depends=['src/simd.hpp.in','src/py_common.hpp','src/render.hpp',
                'src/pyobject.hpp','src/compatibility.hpp'],
            libraries=['SDL']),
        Extension(
            'tracern',
            ['src/tracern.cpp','src/py_common.cpp','src/simd.cpp'],
            depends=['src/simd.hpp.in','src/ntracer_body.hpp',
                'src/py_common.hpp','src/pyobject.hpp','src/geometry.hpp',
                'src/var_geometry.hpp','src/tracer.hpp','src/light.hpp',
                'src/render.hpp','src/camera.hpp','src/compatibility.hpp',
                'src/v_array.hpp','src/instrumentation.hpp'])],
    requires=['pygame'],
    description='A fast hyper-spacial ray-tracing library',
    long_description=long_description,
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: C++',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Multimedia :: Graphics :: 3D Rendering',
        'Topic :: Scientific/Engineering :: Mathematics'],
    cmdclass={
        'build_py' : build_py,
        'build_ext' : CustomBuildExt})
